(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

open Types

type declarationkind =
    Logical
  | Mutable
  | ProcessKind

type declaration = 
    string * (declarationkind * type_expression)

type sk_declaration =
    string * ( type_expression list * declarationkind * type_expression)

(**** renaming of Variables **********************************************)

let rename_decl alpha (varname,blurb) =
  ((try List.assoc varname alpha with Not_found -> varname), blurb)
let rename_dlist alpha = List.map (rename_decl alpha) 

(**** Conversion to Strings **********************************************)
let string_of_declarationkind = function
    Logical -> ""
  | Mutable -> "mutable"
  | ProcessKind -> "process"

let string_of_declaration (varname, (declkind, decltyp)) =
  varname ^ ": " ^ (string_of_declarationkind declkind) ^
    " " ^ (string_of_type_expression decltyp)

let string_of_sk_declaration (varname, (argtypes, declkind, decltyp)) =
  varname ^ 
    (match argtypes with
	 [] -> ""
       | _ -> "(" ^
	   (Utils.string_of_string_list ", "
	      (List.map string_of_type_expression argtypes)) ^
	     ")") ^
    " : " ^ (string_of_declarationkind declkind) ^
    " " ^ (string_of_type_expression decltyp)

let string_of_parlist dl = 
  Utils.string_of_string_list "; "
    (List.map string_of_declaration dl)


(**** Various Transformations *******************************************)
let flatten_declaration strings kind_and_type =
  List.map (function s -> (s, kind_and_type)) strings
let flatten_sk_declaration names_with_types (kind,restype) =
  List.map
    (function (name,types) -> (name,(types,kind,restype)))
    names_with_types

let unskolemize = List.map (function (n,(ts,k,t)) -> (n,(k,t)))

let make_dec_logical = List.map (function (n,t) -> (n,(Logical,t)))


exception Exc_declarations of string
let check_declarations dl =
  begin

    (* check that no variable is declared twice *)
    (try
       ignore(Utils.alldifferent ( List.map (function (x,y) -> x) dl))
     with
	 Utils.Twice x -> raise ( Exc_declarations ("variable \"" ^ x ^
					      "\" declared twice.")));

    (* check that all type expressoins are ground *)
    List.iter
      (function (x,(_,t)) ->
	 if not (is_ground t)
	 then raise (Exc_declarations ("the type expression " ^
				       (string_of_type_expression t) ^
				       " in the declaration of variable " ^
				       x ^ " is not ground.")))
      dl;

    (* check that all type expressions are arrow-free *)
    List.iter
      (function (x,(_,t)) ->
	 if not (arrow_free t)
	 then raise (Exc_declarations ("the type expression " ^
				       (string_of_type_expression t) ^
				       " in the declaration of variable " ^
				       x ^ " is not arrow-free.")))
      dl;

    (* check that all type constructors use only defined type constants *)
    (* and type constructors accoring to their arity.                   *)
    List.iter
      (function (x,(_,t)) ->
	 try
	   check_type_constructors t
	 with
	     Exc_typecons s ->
	       raise (Exc_declarations
			("the type expression in the declaration of " ^
			 "variable " ^ x ^ " " ^ s)))
      dl
  end
      
