(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

{

open Parser
open Roles
open Lexing

(********************************************************************)
(* error reporting                                                  *)
(********************************************************************)

exception Lex_error of string * position

let string_of_lexposition p = 
  "file " ^ p.pos_fname ^
  ", line " ^ (string_of_int p.pos_lnum) ^
  ", column " ^ (string_of_int p.pos_bol)

(*******************************************************************)
(* functions to register update of position                        *)
(*******************************************************************)

(* increment line number *)
let advance_line lexbuf = 
  let oldpos = lexbuf.lex_curr_p
  in
    lexbuf.lex_curr_p <-
    {lexbuf.lex_curr_p with pos_bol = 0; pos_lnum = oldpos.pos_lnum + 1}

(* increment column number *)
let advance_col lexbuf = 
  let oldpos = lexbuf.lex_curr_p
  in
    lexbuf.lex_curr_p <-
    {lexbuf.lex_curr_p with
       pos_bol = oldpos.pos_bol + (String.length (lexeme lexbuf))}

(* set file name *)
let register_filename lexbuf filename =
  lexbuf.lex_curr_p <- {lexbuf.lex_curr_p with pos_fname = filename }
    

(******************************************************************)
(* key words                                                      *)
(******************************************************************)

let keyword_table = Hashtbl.create 30;;
List.iter
  (fun (keyword,token) ->
    Hashtbl.add keyword_table keyword token)
  [
   "role",       ROLE;
   "begin",      BEGIN;
   "end",        END;
   "if",         IF;
   "then",       THEN;
   "else",       ELSE;
   "fi",         FI; 
   "declare",    DECLARE;
   "mutable",    MUTABLE;
   "send",       SEND;
   "recv",       RECV;
   "new",        NEW;
   "match",      MATCH;
   "with",       WITH;
   "case",       CASE;
   "of",         OF;
   "esac",       ESAC;
   "signature",  SIGNATURE;
   "axioms",     AXIOMS;
   "scenario",   SCENARIO;
   "forever",    FOREVER;
   "spawn",      SPAWN;
   "parallel",   PARALLEL;
   "forall",     FORALL;
   "exists",     EXISTS;
   "unique",     UNIQUE;
   "inv",        INV;
   "fail",       FAIL;
   "variables",  VARDECL;
   "choice",     CHOICE;
   "public",     PUBLIC;
   "initial",    INITIAL;
   "always",     ALWAYS;
   "issecret",   ISSECRET;
   "uses",       USES;
	"hashfunctions", HASHFUNCTIONS
 ]

let key_lookup string = 
try 
  Hashtbl.find keyword_table string
with
  Not_found -> IDENT(string)


(*********************************************************************)
(* building tokens                                                   *)
(*********************************************************************)

(* build a STRING token of s without the first and last character *)
(* which are the quotation marks.                                 *)
let stringtoken s =
  (STRING (String.sub s 1 ( (String.length s) - 2 ) ))

}

(*********************************************************************)
(* Lexer rules                                                       *)
(*********************************************************************)
let number = ['0'-'9']
let letter = ['a'-'z''A'-'Z']
let whitespace = ['\r'' ''\t']


rule next_token = parse
  '('    { advance_col lexbuf; PARL }
| ')'    { advance_col lexbuf; PARR }
| "[["   { advance_col lexbuf; DBRAL }
| "]]"   { advance_col lexbuf; DBRAR }
| '['    { advance_col lexbuf; BRAL }
| ']'    { advance_col lexbuf; BRAR }
| '{'    { advance_col lexbuf; ACCL }
| '}'    { advance_col lexbuf; ACCR }
| ":="   { advance_col lexbuf; ASSIGN }
| ';'    { advance_col lexbuf; SC }
| ':'    { advance_col lexbuf; COLON }
| ','    { advance_col lexbuf; COMMA }
| '.'    { advance_col lexbuf; DOT }
| '+'    { advance_col lexbuf; BINOPB("+") }
| '*'    { advance_col lexbuf; STAR }
| '-'    { advance_col lexbuf; BINOPB("-") }
| "::"   { advance_col lexbuf; BINOPB("::") }
| '='    { advance_col lexbuf; BINOPC("=") }
| "=<"   { advance_col lexbuf; BINOPC("=<") }
| ">="   { advance_col lexbuf; BINOPC(">=") }
| '<'    { advance_col lexbuf; BINOPC("<") }
| '>'    { advance_col lexbuf; BINOPC(">") }
| "!="   { advance_col lexbuf; BINOPC("!=") } 
| "&&"   { advance_col lexbuf; BINOPD("&&") }
| "||"   { advance_col lexbuf; BINOPE("||") }
| '|'    { advance_col lexbuf; VBAR }
| '&'    { advance_col lexbuf; AND }
| "->"   { advance_col lexbuf; ARROW }
| "!"    { advance_col lexbuf; NEGATION }
| '@'    { advance_col lexbuf; AT }
| "[-]"  { advance_col lexbuf; BOX }
| letter(letter|number|'_')*
    { advance_col lexbuf; key_lookup(Lexing.lexeme lexbuf) }
| '\'' letter+ { advance_col lexbuf; TYPEVAR(Lexing.lexeme lexbuf) }
| number+
    { advance_col lexbuf; INTEGER(int_of_string(Lexing.lexeme lexbuf)) }
| '\"'[^'\"''\n']*'\"'         { advance_col lexbuf;
                                 stringtoken (Lexing.lexeme lexbuf)}
| whitespace+                  { advance_col lexbuf; next_token lexbuf }
| '\n'                         { advance_line lexbuf; next_token lexbuf }
| '#'[^'\n']*'\n'              { advance_line lexbuf; next_token lexbuf }
| ('#'[^'\n']*)?eof            { THEEND }
| _                        
      { raise (Lex_error(
		 Lexing.lexeme lexbuf, 
		 Lexing.lexeme_start_p lexbuf)) }
 
