(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

(************** abstract sytntax of roles **********************************)

open Expressions
open Types
open Declarations

type rawinstruction =
    IRawLabeled of string * rawinstruction
  | IRawBlock of rawblock
  | IRawSend of expression
  | IRawRecv of pattern
  | IRawFail
  | IRawAssign of string * expression
  | IRawNew of string
  | IRawMatch of expression * pattern
  | IRawConditional of expression * rawinstruction list * rawinstruction list
  | IRawChoice of rawinstruction list list
  | IRawCase of expression * (pattern * rawinstruction list) list
      * rawinstruction list
and rawblock =
    declaration list * rawinstruction list

type instruction =
    IBlock of string*block
  | ISend of string*expression
  | IRecv of string*pattern
  | IFail of string
  | IAssign of string * string * expression
  | INew of string * string
  | IMatch of string * expression * pattern
  | IConditional of string * expression * instruction list * instruction list
  | IChoice of string * (instruction list) list
  | ICase of string *
      expression * (pattern * instruction list) list * instruction list
and block =
    declaration list * instruction list

type role = 
    Role of string * declaration list * block

(*********** processing ***********************************************)

let rec cook_block label (dlist,instlist) =
  (dlist,cook_ilist label instlist)
and cook_ilist label =
   Utils.mapindex (function n -> cook_inst (label^"_"^(string_of_int n)))
and cook_inst label = function
    IRawLabeled(newlabel,inst) -> cook_inst newlabel inst
  | IRawBlock block -> IBlock(label,cook_block label block)
  | IRawSend exp -> ISend(label,exp)
  | IRawRecv pat -> IRecv(label,pat)
  | IRawFail -> IFail(label)
  | IRawAssign(var,exp) -> IAssign(label,var,exp)
  | IRawNew(var) -> INew(label,var)
  | IRawMatch(exp,pat) -> IMatch(label,exp,pat)
  | IRawConditional(cond,thenpart,elsepart) ->
      IConditional(
	label,
	cond,
	(cook_ilist (label^"_thn") thenpart),
	(cook_ilist (label^"_els") elsepart))
  | IRawChoice choicelist ->
      IChoice(
	label,
	(Utils.mapindex 
	   (fun n ilist -> 
	      (cook_ilist (label^"_"^(string_of_int n)) ilist))
	   choicelist))
  | IRawCase(exp,caslist,defaultlist) ->
      ICase(
	label,
	exp,
	(Utils.mapindex 
	   (fun n (pat,ilist) -> 
	      (pat,
	       cook_ilist (label^"_"^(string_of_int n)) ilist))
	   caslist),
	cook_ilist (label^"_dft") defaultlist)
	  
    
let process_instblock labelprefix block = cook_block labelprefix block

(*********** printing *************************************************)

let rec print_block oc prefix (decllist,instlist) =
  output_string oc "declare\n";
  List.iter
    (function decl ->
       output_string oc (prefix^"  ");
       output_string oc (string_of_declaration decl);
       output_string oc ";\n")
    decllist;
  output_string oc prefix;
  output_string oc "begin\n";
  print_instruction_list oc (prefix^"  ") instlist;
  output_string oc prefix;
  output_string oc "end";
and print_instruction_list oc prefix = List.iter
  (function inst -> 
     output_string oc prefix;
     print_instruction oc prefix inst;
     output_string oc ";\n")
and print_instruction oc prefix = function
    ISend (label,exp) ->
      output_string oc (label^": ");
      output_string oc "send(";
      output_string oc (string_of_expression exp);
      output_char oc ')'
  | IRecv (label,pat) ->
      output_string oc (label^": ");
      output_string oc "recv(";
      output_string oc (string_of_pattern pat);
      output_char oc ')'
  | IFail label ->
      output_string oc (label^": ");
      output_string oc "fail"
  | IAssign(label,var,exp) ->
      output_string oc (label^": ");
      output_string oc var;
      output_string oc " := ";
      output_string oc (string_of_expression exp)
  | INew(label,var) ->
      output_string oc (label^": ");
      output_string oc "new(";
      output_string oc var;
      output_char oc ')'
  | IConditional(label,cond,thenpart,elsepart) ->
      output_string oc (label^": ");
      output_string oc "if ";
      output_string oc (string_of_expression cond);
      output_string oc " then\n";
      print_instruction_list oc (prefix^"  ") thenpart;
      output_string oc prefix;
      output_string oc "else\n";
      print_instruction_list oc (prefix^"  ") elsepart;
      output_string oc prefix;
      output_string oc "fi"
  | IMatch(label,exp,pat) ->
      output_string oc (label^": ");
      output_string oc "match ";
      output_string oc (string_of_expression exp);
      output_string oc " with ";
      output_string oc (string_of_pattern pat);
      output_string oc " end"
  | IBlock (label,bl) ->
      output_string oc (label^": ");
      print_block oc prefix bl
  | IChoice(label,choicelist) ->
      output_string oc (label^": ");
      output_string oc "choice\n ";
      List.iter
	(function instl ->
	   output_string oc (prefix^"  |\n");
	   print_instruction_list oc (prefix^"      ") instl;
	   output_char oc '\n')
	choicelist;
      output_string oc prefix;
      output_string oc "end"
  | ICase(label,exp,caselist,elsepart) ->
      output_string oc (label^": ");
      output_string oc "case ";
      output_string oc (string_of_expression exp);
      output_string oc " of ";
      List.iter
	(function (pat,instl) ->
	   output_string oc (prefix^"  ");
	   output_string oc (string_of_pattern pat);
	   output_string oc " ->\n";
	   print_instruction_list oc (prefix^"      ") instl;
	   output_char oc '\n')
	caselist;
      output_string oc prefix;
      output_string oc "else\n";
      print_instruction_list oc (prefix^"  ") elsepart;
      output_string oc prefix;
      output_string oc "fi"



let print_role oc (Role (rolename, decllist, bl)) =
  output_string oc "role ";
  output_string oc rolename;
  output_string oc " (";
  output_string oc (string_of_parlist decllist);
  output_string oc ")\n";
  print_block oc "" bl;
  output_string oc "\n\n"
    
;; 

(************ renaming of variables ********************************)

let rec rename_ilist alpha = List.map (rename_inst alpha)
and rename_inst alpha = function
    IBlock(label,(dl,il)) ->
      let nl = List.map fst dl
	(* newalpha is alpha without the variables declared in dl *)
      in let newalpha =
	  List.filter
	    (function (x,_) -> not (List.mem x nl))
	    alpha
      in
	IBlock(label,(dl, rename_ilist newalpha il))
  | INew (label,x) ->
      (try
	 INew (label,List.assoc x alpha)
       with
	   Not_found -> INew (label,x))
  | IFail label -> IFail label
  | ISend (label,e) -> ISend (label,rename_expr alpha e)
  | IRecv (label,p) -> IRecv (label,rename_pattern alpha p)
  | IAssign (label,x,e) ->
      IAssign (label,
	       (try List.assoc x alpha with Not_found -> x),
	       (rename_expr alpha e))
  | IMatch (label,e,p) -> 
      IMatch (label,
	      (rename_expr alpha e),
	      (rename_pattern alpha p))
  | IConditional (label,cond, thenpart, elsepart) ->
      IConditional (label,
		    (rename_expr alpha cond),
		    (rename_ilist alpha thenpart),
		    (rename_ilist alpha elsepart))
  | IChoice (label,choicelist) ->
      IChoice (label,
	       (List.map (rename_ilist alpha) choicelist))
  | ICase (label,e,caselist,elsepart) ->
      ICase (label,
	     (rename_expr alpha e),
	     (List.map
		(function (p,il) ->
		   (rename_pattern alpha p),
		   (rename_ilist alpha il))
		caselist),
	     (rename_ilist alpha elsepart))
	
(*******************************************************************)

let rec bound_vars_inst = function
    IBlock(_,(dl,il)) -> (List.map fst dl) @ (bound_vars_ilist il)
  | INew _ | ISend _ | IRecv _ | IAssign _ | IFail _ | IMatch _ -> []
  | IConditional(_,_,thenpart,elsepart) ->
      (bound_vars_ilist thenpart) @ (bound_vars_ilist elsepart)
  | IChoice(_,choicelist) ->
      (List.flatten (List.map bound_vars_ilist choicelist))
  | ICase(_,_,caselist,elsepart) ->
      (List.flatten 
	 (List.map (function (p,il) -> bound_vars_ilist il) caselist)) @
	(bound_vars_ilist elsepart)
and bound_vars_ilist il = 
  List.flatten (List.map bound_vars_inst il)
	 

let rec get_form_parlist name = function
    [] -> raise Not_found
  | Role(n,pars,_):: l ->
      if n = name
      then List.map fst pars
      else get_form_parlist name l

let gather_parameter_declarations =
  List.map ( function Role(name,decl_list,_) -> (name,decl_list) )

let gather_toplevel_declarations =
  List.map
    ( function Role(name,decl_list,(block_decl,_)) ->
	(name, decl_list@block_decl) )


(************ collect checkpoints *****************************************)

let rec checkpoints_il il =
  List.fold_right (fun x y -> (checkpoints x)@y) il [] 
and checkpoints = function
    ISend(l,_) | IRecv(l,_) | IFail(l) | IAssign(l,_,_) | INew(l,_) -> [l]
  | IMatch(l,_,_) -> [l]
  | IBlock(l,(_,il)) -> l :: (checkpoints_il il)
  | IConditional(l,_,thenpart,elsepart) ->
      l::(checkpoints_il thenpart)@(checkpoints_il elsepart)
  | IChoice(l,clauselist) -> 
      l::(List.fold_right (fun x y -> (checkpoints_il x)@y) clauselist [])
  | ICase(l,_,clauselist,elselist) ->
      l::(List.fold_right (fun (_,x) y -> (checkpoints_il x)@y) clauselist [])
      @(checkpoints_il elselist)
    
let gather_checkpoints =
  List.map (function Role(name,_,(_,il)) ->
	      (name, ["start"; "end"] @ (checkpoints_il il)))

(*************  collect integer constants  ********************************)

let ic_e = Expressions.int_constants
let ic_el l = List.flatten(List.map ic_e l)
let rec ic_inst = function
    IBlock(_,b) -> ic_block b
  | ISend(_,e) -> ic_e e
  | IRecv(_,l) -> ic_el l
  | IFail _ -> []
  | IAssign(_,_,e) -> ic_e e
  | INew _ -> [] 
  | IMatch(_,e,l) -> (ic_e e) @ (ic_el l)
  | IConditional(_,cond,il_then,il_else) ->
      (ic_e cond) @ (ic_list il_then) @ (ic_list il_else)
  | IChoice(_,choicelist) ->
      (List.flatten (List.map ic_list choicelist))
  | ICase(_,exp,caselist,default) ->
      (ic_e exp) @
      (List.flatten
	 (List.map
	    (function (el,il) -> (ic_el el) @ (ic_list il))
	    caselist)) @
      (ic_list default)
and ic_block b = ic_list (snd b)
and ic_list l =  List.flatten (List.map ic_inst l)
    
let int_constants (Role(name,decl,block)) =
  Utils.singlyfy (ic_block block)

(***************** Typing *************************************************)

exception Exc_roles of string

let rec check_types_block sign env (decl_list, inst_list) =
  (try
     check_declarations decl_list;
     check_types_instlist sign (decl_list@env) inst_list
   with
       Exc_declarations s -> raise
	 (Exc_roles 
	    ("Error in declaration part of a block: " ^ s)))
and check_types_instlist sign env =
  List.iter (check_types_inst sign env)
and check_types_inst sign env = function
    IBlock(_,b) -> check_types_block sign env b
  | ISend(_,e) ->
      (try
	 ignore (type_expr sign env [] [] e)
       with
	   Exc_typeexpression s -> raise 
	     (Exc_roles
		("type error in the argument of a send instruction: " ^ s)))
  | IRecv(_,pl) -> List.iter
      (function e ->
	 (try
	    ignore (type_expr sign env [] [] e)
	  with
	      Exc_typeexpression s -> raise 
		(Exc_roles
		   ("type error in a pattern of of a recv instruction: "
		    ^ s))))
	pl
  | IFail _ -> ()
  | IConditional(_,cond,thenlist,elselist) ->
      (try
	 let guard_type = type_expr sign env [] [] cond
	 in if guard_type <> Typeconst "bool"
	   then raise (Exc_roles
			 ("conditional with guard \"" ^
			  (string_of_expression cond) ^
			  "\" which has type " ^
			  (string_of_type_expression guard_type) ^
			  " instead of bool"))
       with
	   Exc_typeexpression s -> raise
	     (Exc_roles
		("condition \"" ^ (string_of_expression cond) ^
		 "\" can not be typed because: " ^ s)));
      check_types_instlist sign env thenlist;
      check_types_instlist sign env elselist
  | IAssign(_,varname, exp) ->
      (try
	 let vtype = snd (List.assoc varname env)
	 and etype = type_expr sign env [] [] exp
	 in if not (smaller_type etype vtype)
	   then raise (Exc_roles
			 ("assignment of \"" ^ (string_of_expression exp) ^
			  "\" of type " ^ (string_of_type_expression etype) ^
			  " to variable \"" ^ varname ^
			  "\" of type " ^ (string_of_type_expression vtype)))
       with
	   Not_found -> raise (Exc_roles
				 ("assignment to non declared variable \"" ^
				  varname ^ "\""))
	 | Exc_typeexpression s ->raise (Exc_roles
					   ("expression \"" ^ 
					    (string_of_expression exp) ^
					    "\" cannot be typed because "^ s)))
  | INew(_,varname) ->
      (try
	 let vtype = snd (List.assoc varname env)
	 in if not (vtype = Typeconst "nonce"
	       || vtype = Typeconst "int"
	       || vtype = Typeconst "message" )
	   then raise (Exc_roles
			 ("\"new\" instruction on variable \"" ^ varname ^
			  "\" which is of type " ^ 
			  (string_of_type_expression vtype) ^ 
			  "but must be one of \"nonce\" or \"int\""))
       with
	   Not_found -> raise (Exc_roles
				 ("new on a non declared variable \"" ^
				    varname ^ "\"")))
  | IMatch(_,exp, patterns) -> check_types_patternmatch sign env exp patterns
  | IChoice(_,choicelist) ->
      List.iter (check_types_instlist sign env) choicelist 
  | ICase(_,exp,caselist,default) ->
      (* check the types of exp and of the patterns *)
      List.iter
	(function (p,_) -> check_types_patternmatch sign env exp p)
	caselist;
      (* check the types of the instruction lists of all cases *)
      List.iter
	(function (_,il) -> check_types_instlist sign env il)
	caselist;
      (* check the type of the default instruction list *)
      check_types_instlist sign env default
(* (check_types_patternmatch sign env exp p) verifies that, for signature   *)
(* sign and variable declararions env,                                      *)
(* - exp is typable                                                         *)
(* - every pattern in p is typable                                          *)
(* - the type of any pattern p has an infimum with the type of exp          *)
and check_types_patternmatch sign env exp patterns =
  try
    let etype = type_expr sign env [] [] exp
    in List.iter
	 (function p -> 
	    try
	      let pt = type_expr sign env [] [] p
	      in 
		if not (has_inf_type etype pt)
		then raise (Exc_roles
			      ("expression \"" ^
			       (string_of_expression exp) ^
			       "\" has type " ^
			       (string_of_type_expression etype) ^
			       " but is matched against pattern \"" ^
			       (string_of_expression p) ^
			       "\" of type " ^
			       (string_of_type_expression pt)))
	    with
		Exc_typeexpression s ->
		  raise (Exc_roles
			   ( "pattern \"" ^ 
			     (string_of_expression p) ^
			     "\" cannot be typed because " ^ s)))
	 patterns
  with
    | Exc_typeexpression s ->
	raise (Exc_roles
		 ("expression \"" ^ (string_of_expression exp) ^
		  "\" cannot be typed because "^ s))
	

let check_rolename name =
  if List.mem name Types.predefined_type_constants
  then raise
    (Exc_roles ("the name of this role clashes with the name" ^
		  " of a pre-defined type constant."))

(* check a list of roles *)
let check_types_roles sign rolelist=
  List.iter
    (function Role(name,decls,block) ->
       try
	 check_rolename name;
	 check_declarations decls;
	 check_types_block sign decls block
       with
	   Exc_declarations s -> raise
	     (Exc_roles 
		("error in the parameter list of role \"" ^ name ^ "\":\n"
		 ^ s))
	 | Exc_roles s -> raise
	     (Exc_roles
		("error in the body of role \"" ^ name ^ "\":\n" ^ s)))
    rolelist;
  List.iter
    (function (rolename,checkpoint_list) ->
       try
	 Utils.alldifferent checkpoint_list
       with
	   Utils.Twice cp ->
	     raise
	       (Exc_roles
		  ("the checkpoint \"" ^cp^ "\" occurs multiple times " ^
		   "in role \"" ^ rolename ^"\"")))
    (gather_checkpoints rolelist)


    
