(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

open Types
open Intervals
open Expressions
open Declarations


(********* abstract syntax *******************************************)

(* raw abstract syntax *)
type rawscenario =
    SRawLabeled of string * rawscenario
  | SRawCall of string * expression list
  | SRawNew of expression
  | SRawForever of rawscenario list
  | SRawParallel of rawscenario list
  | SRawSequence of rawscenario list
  | SRawAssign of expression * expression
  | SRawConditional of expression * rawscenario list * rawscenario list
  | SRawForall of declaration list * interval list * rawscenario
  | SRawExists of declaration list * rawscenario

(* processed abstract syntax. The first argument is the label of a   *)
(* sub-scenario.                                                     *)
type scenario =
    SCall of string * string * expression list
  | SNew of string * expression
  | SForever of string * scenario list
  | SParallel of string * scenario list
  | SSequence of string * scenario list
  | SAssign of string * expression * expression
  | SConditional of string * expression * scenario list * scenario list
  | SForall of string * declaration list * interval list * scenario
  | SExists of string * declaration list * scenario


(*************** processing *******************************************)

let rec process_scenario label = function
    SRawLabeled(newlabel,scene) -> process_scenario newlabel scene
  | SRawCall(procname,args) -> SCall(label,procname,args)
  | SRawNew(exp) -> SNew(label,exp)
  | SRawForever(instlist) -> SForever(label, process_list label instlist)
  | SRawParallel(instlist) -> SParallel(label, process_list label instlist)
  | SRawSequence(instlist) -> SSequence(label, process_list label instlist)
  | SRawAssign(lexp,rexp) -> SAssign(label,lexp,rexp)
  | SRawConditional(cond,thenpart,elsepart) ->
      SConditional(label,cond,
		   (process_list (label^"_thn") thenpart),
		   (process_list (label^"_els") elsepart))
  | SRawExists(decllist,scene) ->
      SExists(label, decllist, process_scenario (label^"_bdy") scene)
  | SRawForall(decllist,intervals,scene) ->
      SForall(label, decllist, intervals,
	      process_scenario (label^"_bdy") scene)
and process_list label =
  Utils.mapindex (function n ->  process_scenario(label^"_"^(string_of_int n)))

(*************** printing *********************************************)

(* (print_sc oc prefix inst) prints to the output channel oc the instruction *)
(* inst, possibly on several lines, terminated by a "\n". "prefix" is the    *)
(* indention of lines.                                                       *)
let rec print_sc oc prefix = function
    SCall(label,rulename, args) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc rulename;
      output_char oc '(';
      output_string oc (string_of_expression_list args);
      output_string oc ")"
  | SNew(label,exp) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "new";
      output_char oc '(';
      output_string oc (string_of_expression exp);
      output_string oc ")"
  | SAssign(label,lexp,rexp) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc (string_of_expression lexp);
      output_string oc " := ";
      output_string oc (string_of_expression rexp);
 | SConditional(label,cond,thenpart,elsepart) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "if ";
      output_string oc (string_of_expression cond);
      output_string oc " then\n";
      print_sc_list oc (prefix^"  ") thenpart;
      output_string oc prefix;
      output_string oc "else\n";
      print_sc_list oc (prefix^"  ") elsepart;
      output_string oc prefix;
      output_string oc "fi"
 | SForever (label,sl) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "forever\n";
      print_sc_list oc (prefix^"  ") sl;
      output_string oc prefix;
      output_string oc "end";
 | SSequence (label,sl) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "begin\n";
      print_sc_list oc (prefix^"  ") sl;
      output_string oc prefix;
      output_string oc "end";
 | SParallel (label,sl) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "parallel\n";
      print_sc_par_list oc (prefix^"  ") sl;
      output_string oc "end";
 | SExists (label,dl,inst) ->
     output_string oc prefix;
     output_string oc (label^": ");
     output_string oc "exists ";
     output_string oc (string_of_parlist dl);
     output_string oc " .\n";
     print_sc oc (prefix^"  ") inst;
 | SForall (label,dl,intervals,inst) ->
     output_string oc prefix;
     output_string oc (label^": ");
     output_string oc "forall ";
     output_string oc (string_of_parlist dl);
     output_string oc (string_of_interval_list intervals);
     output_string oc " .\n";
     print_sc oc (prefix^"  ") inst;

(* (print_sc_list outchannel prefix sclist) outputs to                *)
(* the output channel oc the list of scene instructions sclist, where *)
(* every instruction is prefixed by prefix and terminated by ";\n"     *)
and print_sc_list oc prefix =
  List.iter
    (function inst ->
       print_sc oc prefix inst;
       output_string oc ";\n")
    
(* (print_sc_list outchannel prefix sclist) outputs to                 *)
(* the output channel oc the list of scene instructions sclist, where  *)
(* every instruction is prefixed by prefix^"  " and terminated by "\n" *)
(* except for the first instruction which is prefixed by prefix^"| "   *)
and print_sc_par_list oc prefix = function
    [] -> ()
  | h::r ->
      print_sc oc (prefix^"  ") h;
      output_string oc "\n";
      List.iter
	(function inst ->
	   output_string oc (prefix^"|\n");
	   print_sc oc (prefix^"  ") inst;
	   output_string oc "\n")
	r
    

let print_scenario oc = print_sc oc ""

(*************** extract integer constants ****************************)
let rec ic = function 
    SCall(_,_,el) -> List.flatten (List.map Expressions.int_constants el)
  | SForever (_,l) | SSequence (_,l) | SParallel (_,l) -> ic_list l
  | SAssign(_,le,re) ->
      (Expressions.int_constants le) @ (Expressions.int_constants re)
  | SNew(_) -> []
  | SConditional(_,cond,thenl,elsel) ->
      (Expressions.int_constants cond) @
      (ic_list thenl) @ (ic_list elsel)
  | SForall(_,_,_,s) | SExists(_,_,s) -> ic s
and ic_list l = List.flatten (List.map ic l)

let int_constants l = Utils.singlyfy (ic l)


(*************** collect checkpoints *********************************)

let rec checkpoints = function
    SCall(l,_,_) | SNew(l,_) | SAssign(l,_,_) -> [l]
  | SForever(l,slist) | SParallel(l,slist) | SSequence(l,slist) ->
      l::(checkpoints_list slist)
  | SExists(l,_,scene) | SForall(l,_,_,scene) ->
      l::(checkpoints scene)
  | SConditional(l,_,thenpart,elsepart) ->
      l::(checkpoints_list thenpart)@(checkpoints_list thenpart)
and checkpoints_list slist = 
  List.fold_right (fun x y -> (checkpoints x)@y) slist []

let gather_checkpoints scenario = "start"::"end"::(checkpoints scenario)

(*************** typing **********************************************)

exception Exc_scenario of string
let rec type_scenario sign role_decl gl_var_decl env = function
    SCall(_,"makekey",arguments) ->
      (match arguments with
	   [alg_arg; pkey_arg] ->
	     let alg_t = 
	       try type_expr sign env gl_var_decl [] alg_arg
	       with
		   Exc_typeexpression s -> 
		     raise (Exc_scenario
			      ("cannot type first argument \"" ^
			       (string_of_expression alg_arg) ^
			       "\" of makekey because " ^ s))
	     and pkey_t =
	       try type_expr sign env gl_var_decl [] pkey_arg
	       with
		   Exc_typeexpression s -> 
		     raise (Exc_scenario
			      ("cannot type second argument \"" ^
			       (string_of_expression pkey_arg) ^
			       "\" of makekey because " ^ s))
	     in
	       (match alg_t with
		    Typeconst "algo" ->
		      if pkey_t <> Typeconst "privkey"
		      then raise (Exc_scenario
				    ("second argument \"" ^
				       (string_of_expression pkey_arg) ^
				       "\" of makekey has type " ^
				       (string_of_type_expression pkey_t) ^
				       " instead of privkey"))
		  | Typeconst "symalgo" ->
		      if pkey_t <> Typeconst "symkey"
		      then raise (Exc_scenario
				    ("second argument \"" ^
				       (string_of_expression pkey_arg) ^
				       "\" of makekey has type " ^
				       (string_of_type_expression pkey_t) ^
				       " instead of symkey"))
		  | _ -> raise (Exc_scenario
			     ("first argument \"" ^
				(string_of_expression alg_arg) ^
				"\" of makekey has type " ^
				(string_of_type_expression alg_t) ^
				" instead of algo or symalgo")))

	 | _ -> raise (Exc_scenario
			 ("makekey needs 2 arguments but is called with " ^
			  (string_of_int (List.length arguments)) ^
			  " arguments")))
  | SCall(_,rolename,arguments) ->
      (try
	 let form_parameters = List.assoc rolename role_decl
	 in
	 let fl = List.length form_parameters
	 and al = List.length arguments
	 in if fl <> al
	   then raise
	     (Exc_scenario
		("\"" ^ rolename ^ "\" is declared with " ^
		 (string_of_int fl) ^
		 " formal parameters but is here instantiated with " ^
		 (string_of_int al) ^
		 " actual arguments"))
	   else  List.iter2 
	     (fun (formname, (_,formtype)) expr ->
		try
		  let acttype = type_expr sign env gl_var_decl [] expr
		  in
		    if not (smaller_type acttype formtype)
		    then raise
		      (Exc_scenario
			 ("instantiaton of role \"" ^ rolename ^
			  "\": formal parameter \"" ^ formname ^
			  "\" is declared with type " ^
			  (string_of_type_expression formtype) ^
			  " but is instantiated with the expression \""^
			  (string_of_expression expr) ^
			  "\" which is of type "^
			  (string_of_type_expression  acttype)))
		with
		    Exc_typeexpression s -> raise (Exc_scenario s))
	     form_parameters
	     arguments
       with
	   Not_found -> raise
	     (Exc_scenario
		("instantiaton of a role \"" ^ rolename ^
		 "\" which is not declared")))
  | SForever (_,l) | SParallel (_,l) | SSequence (_,l) ->
      List.iter (type_scenario sign role_decl gl_var_decl env) l
  | SForall(_,decl_list, interval_list, body) ->
      check_declarations decl_list;
      check_interval_list interval_list;
      if ( List.length interval_list > 0 &&
	    List.length decl_list <> List.length interval_list)
      then raise (Exc_scenario
		    ("the number of variables (" ^ 
		       (string_of_int (List.length decl_list)) ^
		       ") and of intervals (" ^
		       (string_of_int (List.length interval_list)) ^
		       ") in a forall do not correspond"));
      type_scenario sign role_decl gl_var_decl (decl_list@env) body
  | SExists(_,decl_list, body) ->
      check_declarations decl_list;
      type_scenario sign role_decl gl_var_decl (decl_list@env) body
  | SConditional(_,cond,thenlist,elselist) ->
      (try
	 let guard_type = type_expr sign env gl_var_decl [] cond
	 in if guard_type <> Typeconst "bool"
	   then raise (Exc_scenario
			 ("conditional with guard \"" ^
			  (string_of_expression cond) ^
			  "\" which has type " ^
			  (string_of_type_expression guard_type) ^
			  " instead of bool"))
       with
	   Exc_typeexpression s -> raise
	     (Exc_scenario
		("condition \"" ^ (string_of_expression cond) ^
		 "\" can not be typed because: " ^ s)));
      type_scenlist sign role_decl gl_var_decl env thenlist;
      type_scenlist sign role_decl gl_var_decl env elselist
  | SAssign(_,lexp, exp) ->
      if not (is_l_expression lexp)
      then raise (Exc_scenario
		    ("assignment to \"" ^ (string_of_expression lexp) ^
		     "\" which is not an l-expression"));
      let vtype = try
	type_expr sign env gl_var_decl [] lexp
      with
	  Exc_typeexpression s -> 
	    raise (Exc_scenario
		     ("expression \"" ^ 
		      (string_of_expression lexp) ^
		      "\" cannot be typed because "^ s))
      and etype = try
	type_expr sign env gl_var_decl [] exp
      with
	  Exc_typeexpression s -> 
	    raise (Exc_scenario
		     ("expression \"" ^ 
		      (string_of_expression exp) ^
		      "\" cannot be typed because "^ s))
      in if not (smaller_type etype vtype)
	then raise (Exc_scenario
		      ("assignment of \"" ^ (string_of_expression exp) ^
		       "\" of type " ^ (string_of_type_expression etype) ^
		       " to variable \"" ^ (string_of_expression lexp) ^
		       "\" of type " ^ (string_of_type_expression vtype)))
  | SNew(_,lexp) ->
      if not (is_l_expression lexp)
      then raise (Exc_scenario
		    ("\"new\" on \"" ^ (string_of_expression lexp) ^
		     "\" which is not an l-expression"));
      let vtype = try
	type_expr sign env gl_var_decl [] lexp
      with
	  Exc_typeexpression s -> 
	    raise (Exc_scenario
		     ("expression \"" ^ 
			(string_of_expression lexp) ^
		      "\" cannot be typed because "^ s))
      in  if not (vtype = Typeconst "nonce" 
	    || vtype = Typeconst "int" 
	    || vtype = Typeconst "message")
	then raise (Exc_scenario
		      ("\"new\" instruction on expression \"" ^
			 (string_of_expression lexp) ^
			 "\" which is of type " ^ 
			 (string_of_type_expression vtype) ^ 
			 " but must be one of \"nonce\" or \"int\""))
and type_scenlist sign role_decl gl_var_decl env =
  List.iter (type_scenario sign role_decl gl_var_decl env)

let check_types_scenario sign role_decl gl_var_decl sceninst =
  type_scenario sign role_decl gl_var_decl
    (Declarations.unskolemize gl_var_decl) sceninst;
  try
    Utils.alldifferent (gather_checkpoints sceninst)
  with
      Utils.Twice cp ->
	raise
	  (Exc_scenario
	     ("the checkpoint \"" ^cp^ "\" occurs multiple times " ^
		"in the scenario"))
    

